# -*- coding: utf-8 -*-
"""Specifies the version of the Auto_TS package."""

__title__ = "Auto_TS"
__author__ = "Ram Seshadri"
__description__ = "Build time series models for any data set, any size. Now using dask."
__url__ = "https://github.com/Auto_ViML/Auto_TS.git"
__version__ = "0.0.64"
__license__ = "Apache License 2.0"
__copyright__ = "2020-22 Google"
